// b64tobin
//    Jim Russell, SynData Technologies, Inc.

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>


typedef unsigned char   UCHAR;
typedef int             SYN_RC;
UCHAR vcBase64[] =
   "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

SYN_RC   SynBase64(char*, char*, bool);
SYN_RC   StreamB64Encode(istream& is, ostream& os, bool bBreakLines);
SYN_RC   StreamB64Decode(istream& is, ostream& os);

int
main(
   int      argc,
   char**   argv)
{
   if (argc < 3)
   {
      printf("Usage: B64TOBIN <infile> <outfile>\n");
      exit(1);
   }
   return SynBase64(argv[1], argv[2], true);
}

//****************************************************************************
SYN_RC
SynBase64(
   char *   zFspecSrc,
   char *   zFspecDst,
   bool     bDecoding)
{
   fstream  fSrc;
   fstream  fDst;

   fSrc.open(zFspecSrc, ios::in | ios::binary);
   if (fSrc.fail())  return -1;

   fDst.open(zFspecDst, ios::out | ios::binary);
   if (fDst.fail())  return -2;


   if (bDecoding)
      return StreamB64Decode(fSrc, fDst);
   else
      return StreamB64Encode(fSrc, fDst, true);
}
//****************************************************************************
SYN_RC                            // returns bytes output
StreamB64Encode(
   istream& is,
   ostream& os,
   bool     bBreakLines)
{
   UCHAR    cEndMode;
   UCHAR    cIn[3];
   UCHAR    wIxLine;

   wIxLine = 0;                     // write 76 chars per line

   for (;;)
   {
      is.get(cIn[0]);   if (!is) { cEndMode = 0; break; }
      is.get(cIn[1]);   if (!is) { cEndMode = 1; break; }
      is.get(cIn[2]);   if (!is) { cEndMode = 2; break; }

      os.put(vcBase64[cIn[0]>>2]);
      os.put(vcBase64[((cIn[0]&'\x03')<<4) | (cIn[1]>>4)]);
      os.put(vcBase64[((cIn[1]&'\x0F')<<2) | (cIn[2]>>6)]);
      os.put(vcBase64[cIn[2]&'\x3F']);

      if (bBreakLines)
      {
         wIxLine+=4;
         if (wIxLine > 72)
         {
            os.put('\x0D');
            os.put('\x0A');
            wIxLine = 0;
         }
      }
   }

   switch (cEndMode)
   {
   case 0:
      break;
   case 1:
      cIn[1] = 0;
      cIn[2] = 0;
      os.put(vcBase64[cIn[0]>>2]);
      os.put(vcBase64[((cIn[0]&'\x03')<<4) | (cIn[1]>>4)]);
      os.put('=');
      os.put('=');
      break;
   case 2:
      cIn[2] = 0;
      os.put(vcBase64[cIn[0]>>2]);
      os.put(vcBase64[((cIn[0]&'\x03')<<4) | (cIn[1]>>4)]);
      os.put(vcBase64[((cIn[1]&'\x0F')<<2) | (cIn[2]>>6)]);
      os.put('=');
      break;
   }

   if (bBreakLines & wIxLine)             // unfinished line?
   {
      os.put('\x0D');
      os.put('\x0A');
   }

   if (!os)
      return -1;

   return 0;
}
//****************************************************************************
SYN_RC
StreamB64Decode(
   istream& is,
   ostream& os)
{
   UCHAR    cIn1;
   UCHAR    cIn2;
   UCHAR    cIn3;
   UCHAR    cIn4;
   UCHAR    cBuf[2];

   cBuf[1] = 0;   // stupid pet trick for strcspn (Scan vcBase64 for a char)

   while (is)
   {
      is.get(cBuf[0]);
      if (!is)    break;

      if ('\x0D' == cBuf[0] || '\x0A' == cBuf[0])
      {
         is.get(cBuf[0]);
         if (!is)    break;
         continue;
      }

      cIn1 = (UCHAR) strcspn((char*) vcBase64, (char*) cBuf);

      is.get(cBuf[0]);
      if (!is)    break;
      cIn2 = (UCHAR) strcspn((char*) vcBase64, (char*) cBuf);

      os.put((UCHAR) ((cIn1 << 2) | (cIn2 >> 4)));

      is.get(cBuf[0]);
      if (!is)    break;

      if ('=' == cBuf[0])
      {
         is.get(cBuf[0]);
         if (!is)    break;
         if ('=' == cBuf[0])   // which it should...
            ;
         else
            return -1;
      }
      else
      {
         cIn3 = (UCHAR) strcspn((char*) vcBase64, (char*) cBuf);
         os.put((UCHAR) ((cIn2 << 4) | (cIn3 >> 2)));
         is.get(cBuf[0]);
         if (!is)    break;
         if ('=' == cBuf[0])
         {
            // shouldn't have to do anything here
         }
         else
         {
            cIn4 = (UCHAR) strcspn((char*) vcBase64, (char*) cBuf);
            os.put((UCHAR) ((cIn3 << 6) | (cIn4 & '\x3F')));
         }
      }
   }
   if (!os)
      return -1;

   return 0;
}
//****************************************************************************
//****************************************************************************
//****************************************************************************
