// brute.cpp
//    Shell for Counterpane S/MIME searcher
//    v1.0  Oct 1997 Jim Russell, SynData Technologies
//
//    This module serves as the main module for the Win32 console 
//    application BRUTE.EXE, and also as the search engine for the
//    screen saver BRUTESAV.SCR, based on whether BRUTECON or BRUTSAV
//    is defined by the compiler.
//
//    I've taken the guts of Chris Hall's 'search.c', and embedded it
//    in here as the function search().
//
//****************************************************************************

#include <direct.h>
#include <process.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(BRUTESAV)
#include <windows.h>
#endif

#include <iostream.h>
#include <fstream.h>

typedef unsigned char   Uchar;
typedef unsigned long   Ulong;
typedef unsigned short  Ushort;
typedef void (*fnSignal)(int);

extern "C"
{

#include "rc2.h"

#if defined(BRUTESAV)                  // In the Screen Saver version,
extern void    search_one_chunk(void); // this is called by C code elsewhere
extern short   iCurrCycle;             // and these are referenced externally
extern char    hexNextKey[];
extern char    vcMsgBuf[];
extern short   bEnabled;
#endif

}
                                       // local prototypes
static int     search(Uchar*, Uchar*, Ulong);
static void    wrapup(void);
static void    HexToBin(char*, Uchar*, short);
static void    BinToHex(Uchar*, char*, short);
static void    GetState(char*);
static void    PutState(char*);
static void    PutResults(char*, char*);
static void    LoadCtext(char*);
static void    FreeCtext(void);


                                       // state of the search
short    iCurrCycle   = 0;             // cycle id [0-27]
char     fspecCtext[121];              // filespec of *parsed* ciphertext
char     hexNextKey[11];               // next key to search - ascii hex
Uchar    NextKey[5];                   // next key to search - real binary
short    bEnabled = true;              // this lets the screen saver know
                                       //    if the search is alive


struct Cycle
{
   Ulong    uCycLenHi;     // cyclen / 2^20
   Ulong    uCycLenLo;     // cyclen % 2^20
   Uchar    DelimKey[5];   // a key in the cycle, use to delimit it
};

const Ulong  k2to20th = 0x00100000ul;       // 1048576;

Cycle cyctable[]
=  {
      {0x00000ul, 0x00001ul, {0xFE, 0xFE, 0xFE, 0xFE, 0xFE}},  // 0
      {0x00000ul, 0x0000Eul, {0x07, 0x74, 0x38, 0xD1, 0xF8}},
      {0x00000ul, 0x00011ul, {0x17, 0x8C, 0x6B, 0x98, 0xAA}},
      {0x00000ul, 0x00026ul, {0x02, 0x4F, 0xE2, 0x24, 0x7E}},
      {0x00000ul, 0x002D2ul, {0x00, 0x3F, 0x0B, 0xBD, 0x43}},
      {0x00000ul, 0x00313ul, {0x00, 0x02, 0xD4, 0x93, 0xAB}},  // 5
      {0x00000ul, 0x00BBBul, {0x00, 0x07, 0x47, 0x36, 0x9B}},
      {0x00000ul, 0x00FBAul, {0x00, 0x0C, 0xA5, 0x2E, 0x69}},
      {0x00000ul, 0x05310ul, {0x00, 0x04, 0xD2, 0x3B, 0xB4}},
      {0x00000ul, 0x055D2ul, {0x00, 0x01, 0x25, 0x3E, 0xDF}},
      {0x00000ul, 0x144B7ul, {0x00, 0x02, 0xF7, 0x53, 0x6A}},  // 10
      {0x00000ul, 0x2893Bul, {0x00, 0x00, 0xE6, 0x8E, 0x6D}},
      {0x00000ul, 0x42DFFul, {0x00, 0x00, 0x17, 0xF9, 0xE9}},

      {0x00001ul, 0x06852ul, {0x00, 0x00, 0x0a, 0x96, 0x77}},
      {0x00001ul, 0x252CBul, {0x00, 0x00, 0x29, 0xf7, 0x0c}},
      {0x00001ul, 0x86D24ul, {0x00, 0x00, 0x06, 0x61, 0x67}},  // 15
      {0x00002ul, 0x1DBCEul, {0x00, 0x00, 0x0e, 0x58, 0xbb}},
      {0x00024ul, 0xE8F3Ful, {0x00, 0xcb, 0x82, 0x6a, 0xee}},
      {0x0007Dul, 0x13D5Dul, {0x00, 0x00, 0x00, 0x2d, 0x03}},
      {0x00082ul, 0xD5067ul, {0x00, 0x00, 0x00, 0x3e, 0xa8}},
      {0x000F3ul, 0xEBB1Aul, {0x00, 0x00, 0x00, 0x05, 0x8b}},  // 20
      {0x002B2ul, 0xA6AA4ul, {0x00, 0x00, 0x00, 0x03, 0x91}},
      {0x00364ul, 0x4D184ul, {0x00, 0x00, 0x00, 0x01, 0x87}},
      {0x0202Ful, 0x7C3FBul, {0x00, 0x00, 0x00, 0x00, 0x0d}},
      {0x05FF4ul, 0xFA427ul, {0x00, 0x00, 0x00, 0x00, 0x21}},
      {0x19739ul, 0xFC8B8ul, {0x00, 0x00, 0x00, 0x00, 0x05}},  // 25
      {0x439BCul, 0x6A73Cul, {0x00, 0x00, 0x00, 0x00, 0x06}},
      {0x9A6AFul, 0x14923ul, {0x00, 0x00, 0x00, 0x00, 0x00}}   // 27
   };

FILE*    fpCtext;    // parsed Ciphertext file
Uchar    IV[8];
long     lenCtext;
Uchar*   plaintext;
Uchar*   ciphertext;
Ulong    ulTrials;
char     key[5];
char     keyhex[11];
char     vcTrials[33];
char     vcCmdLine[120];
char     fspecState[] = "brutests.txt";
char     fspecResults[] = "brutelog.txt";
//char     fspecCtext[] = "bruterc2.bin";
char     vcMsgBuf[1024];

//****************************************************************************
inline Uchar
hex(char c)
{
   if (c >= '0' && c <= '9')
      return c - '0';
   if (c >= 'A' && c <= 'F')
      return c - 'A' + 10;
   if (c >= 'a' && c <= 'f')
      return c - 'a' + 10;
   return 0xFF;
}
//****************************************************************************
inline void
xorbuf(Ushort *d, Ushort *s)
{
   d[0] ^= s[0];
   d[1] ^= s[1];
   d[2] ^= s[2];
   d[3] ^= s[3];
}
//****************************************************************************
inline bool
is_text(Uchar* p)
{
   if (p[0] >= 128)  return false;
   if (p[1] >= 128)  return false;
   if (p[2] >= 128)  return false;
   if (p[3] >= 128)  return false;
   if (p[4] >= 128)  return false;
   if (p[5] >= 128)  return false;
   if (p[6] >= 128)  return false;
   if (p[7] >= 128)  return false;
   return true;
}
//****************************************************************************
inline void
show_msg(
   char* msg)
{
#if defined(BRUTECON)
   printf("\n%s", msg);
#endif
}
//****************************************************************************
inline void
exit_w_msg(
   char* msg,
   int   errcode = -1)
{
   show_msg(msg);
   PutState(fspecState);
   FreeCtext();
#if defined(BRUTECON)
   exit(errcode);
#elif defined(BRUTESAV)
   bEnabled = false;
#endif
}
//****************************************************************************
int
search(
   Uchar* StartKey,                    // key to start search cycle
   Uchar* EndKey,                      // key that triggers end of search
   Ulong  uMaxTrials = 0)              // max number of trials (0 = no max)
{
                                       /*
                                        * odd_offset_key has 259 elements
                                        * so we have a dummy slot at the end
                                        * and two dummy slots at the front.
                                        */
   Uchar    odd_offset_key[257];
   Uchar    even_offset_key[257];

   Uchar*   end;     // not really used yet
   Uchar*   oo;
   Uchar*   eo;
   Uchar*   oe;
   Uchar*   ee;
   Uchar    key_0[5];
   Ushort*  p;       // an alias for plaintext
   Ushort*  c;       // an alias for ciphertext
   Ushort*  tp;
   Ushort*  tc;
   short    match;


   c = (Ushort*)ciphertext;
   p = (Ushort*)plaintext;

   memcpy(key_0, StartKey, 5);

   /* need to have two different key schedules since two successive key
    * key schedules have different alignments.  the two data structures
    * will make it so that the current key schedule always has proper
    * word alignment so we can cast to a (Ushort *).
    */

   memset(odd_offset_key,  0, sizeof(odd_offset_key));
   memset(even_offset_key, 0, sizeof(even_offset_key));

   /* oo and eo point to the current key schedule.  they effectively
    * represent a sliding window through odd_offset_key and
    * even_offset_key respectively.  once oo and eo pass through oe
    * and ee respectively, then they are reset to just past the
    * beginning of their respective buffers.  as a result we must
    * always make sure that we update the bytes in the lower half
    * of the buffers so that they will be correct when we reset oo
    * and eo.
    */

   oo = odd_offset_key + 2;
   eo = even_offset_key + 2;
   oe = oo + 128;
   ee = eo + 128;

   memcpy(eo, key_0, sizeof(key_0));
   rc2_full_keyschedule(eo);
   memcpy(oo, eo+1, 127);
   rc2_partial_keyschedule(oo);

   end = ciphertext + lenCtext;
                                       // perform trials until we hit a high-
                                       // probability match, or get to the
                                       // "delimiter" key
   for (Ulong ul=0; /*no test*/; ul++)
   {
      /* save NextKey for restarting */
      NextKey[0] = inverse[eo[0]];
      NextKey[1] = eo[1];
      NextKey[2] = eo[2];
      NextKey[3] = eo[3];
      NextKey[4] = eo[4];

      /* perform a trial encryption with the key in even_offset_key */
      rc2_decrypt((Ushort*)eo, c, p);
      xorbuf(p, (Ushort *)IV);
      for (tp = p, tc = c, match = 0; match < 5; match++)
      {
         if (is_text((Uchar*)tp) == 0)
            goto fail_1;
         tp += 4;
         tc += 4;
         rc2_decrypt((Ushort*)eo, tc, tp);
         xorbuf(tp, tc-4);
      }


      /* almost definitely found a match! */
      *(tp+4) = 0;
      if (strstr((char*)p, "Content-") != NULL)
      {
         sprintf(vcMsgBuf,
            "Very high probability match.\n"
            "key = %02x%02x%02x%02x%02x.\n"
            "first 40 bytes of plaintext.\n%s\n",
            inverse[eo[0]], eo[1], eo[2], eo[3], eo[4], plaintext);
         PutResults(fspecResults, vcMsgBuf);
         PutState(fspecState);
         exit_w_msg(vcMsgBuf, 1);
      }
      else
      {
         sprintf(vcMsgBuf,
            "High probability match.\n"
            "key = %02x%02x%02x%02x%02x.\n"
            "first 40 bytes of plaintext.\n%s\n",
            inverse[eo[0]], eo[1], eo[2], eo[3], eo[4], plaintext);
         PutResults(fspecResults, vcMsgBuf);
         PutState(fspecState);
         show_msg(vcMsgBuf);
      }

fail_1:

      /* the next key for even_offset_key comes two iterations later in
       * the cycle than the current one, *not* one.  so get the info on
       * the next iteration from odd_offset_key and compute the second
       * iteration.
       */
      eo[0] = eo[128] = oo[127];

      /* check if we need to wrap back to the beginning of the buffer */
      if ((eo += 2) == ee)
         eo = even_offset_key + 2;
      rc2_partial_keyschedule(eo);

      /* update in the event we wrap back to the beginning */
      eo[-1] = eo[127];

                                       // if the next key we are going
                                       // to do is our delimiter, exit
      if (  oo[1] == EndKey[1]
         && oo[2] == EndKey[2]
         && oo[3] == EndKey[3]
         && oo[4] == EndKey[4]
         && inverse[oo[0]] == EndKey[0]  )
      {
         PutState(fspecState);
         return 0;
      }

      /* save NextKey for restarting */
      NextKey[0] = inverse[oo[0]];
      NextKey[1] = oo[1];
      NextKey[2] = oo[2];
      NextKey[3] = oo[3];
      NextKey[4] = oo[4];

      /* perform a trial decryption with the key in odd_offset_key */
      rc2_decrypt((Ushort*)oo, c, p);
      xorbuf(p, (Ushort *)IV);
      for (tp = p, tc = c, match = 0; match < 5; match++)
      {
         if (is_text((Uchar*)tp) == 0)
            goto fail_2;
         tp += 4;
         tc += 4;
         rc2_decrypt((Ushort*)oo, tc, tp);
         xorbuf(tp, tc-4);
      }

      /* almost definitely found a match! */
      *(tp+4) = 0;
      if (strstr((char*)p, "Content-") != NULL)
      {
         sprintf(vcMsgBuf,
            "Very high probability match.\n"
            "key = %02x%02x%02x%02x%02x.\n"
            "first 40 bytes of plaintext.\n%s\n",
            inverse[oo[0]], oo[1], oo[2], oo[3], oo[4], plaintext);
         PutResults(fspecResults, vcMsgBuf);
         PutState(fspecState);
         exit_w_msg(vcMsgBuf, 1);
      }
      else
      {
         sprintf(vcMsgBuf,
            "High probability match.\n"
            "key = %02x%02x%02x%02x%02x.\n"
            "first 40 bytes of plaintext.\n%s\n",
            inverse[oo[0]], oo[1], oo[2], oo[3], oo[4], plaintext);
         PutResults(fspecResults, vcMsgBuf);
         PutState(fspecState);
         show_msg(vcMsgBuf);
         /* keep on searching */
      }

fail_2:

      /* the next key for odd_offset_key comes two iterations later in
       * the cycle than the current one, *not* one.  so get the info on
       * the next iteration from even_offset_key and compute the second
       * iteration.
       */
      oo[0] = oo[128] = eo[127];

      /* check if we need to wrap back to the beginning of the buffer */
      if ((oo += 2) == oe)
         oo = odd_offset_key + 2;
      rc2_partial_keyschedule(oo);

      /* update in the event we wrap back to the beginning */
      oo[-1] = oo[127];

                                       // if the next key we are going
                                       // to do is our delimiter, exit
      if (  eo[1] == EndKey[1]
         && eo[2] == EndKey[2]
         && eo[3] == EndKey[3]
         && eo[4] == EndKey[4]
         && inverse[eo[0]] == EndKey[0]  )
      {
         NextKey[0] = inverse[eo[0]];
         NextKey[1] = eo[1];
         NextKey[2] = eo[2];
         NextKey[3] = eo[3];
         NextKey[4] = eo[4];

         PutState(fspecState);
         return 0;
      }

      if (uMaxTrials)
      {
         if (ul >= uMaxTrials)
         {
            PutState(fspecState);
            return 2;
         }
      }
      else
      if (ul > k2to20th)
      {
#if defined(BRUTECON)
         printf(".");
#endif
         ul=0;
         PutState(fspecState);
      }
   }
}
//****************************************************************************
void 
wrapup(void)
{
   exit_w_msg("\nExiting search.", 0);
}
//****************************************************************************
void
HexToBin(
   char*    pHex,
   Uchar*   pBin,
   short    len)
{
   for (short i=0; i<len; i++)
      pBin[i] = (hex(pHex[2*i]) << 4) | hex(pHex[2*i+1]);
}
//****************************************************************************
void
BinToHex(
   Uchar*   pBin,
   char*    pHex,
   short    inlen)
{
   char     zTmp[3] = "";
   int      ix;

	for (ix=0; ix < inlen; ix++)
   {
      itoa((Uchar)pBin[ix], zTmp, 16);
      if (zTmp[1] == 0)
      {
      	zTmp[1] = zTmp[0];
         zTmp[0] = '0';
      }
   	memcpy(pHex + (ix * 2), zTmp, 2);
   }

   pHex[(ix * 2)] = 0;
   return;
}
//****************************************************************************
void
GetState(
   char*    fspec)
{
   ifstream    fIn(fspec);
   fIn >> iCurrCycle;
   fIn >> fspecCtext;
   fIn >> hexNextKey;
   if (fIn.fail())
   {
      sprintf(vcMsgBuf, "\nNo %s found.\n", fspec);
      iCurrCycle = 0;
      fspecCtext[0] = 0;
      hexNextKey[0] = 0;
      exit_w_msg(vcMsgBuf);
   }
   fIn.close();
                                       // convert the hex ascii to binary
   HexToBin(hexNextKey, NextKey, 5);
}
//****************************************************************************
void
PutState(
   char*    fspec)
{
   if (*fspecCtext)
   {
      ofstream    fOut(fspec);
      BinToHex(NextKey, hexNextKey, 5);
      fOut << iCurrCycle << " " << fspecCtext << " " << hexNextKey << "\n";
   }
}
//****************************************************************************
void
PutResults(
   char*    fspec,
   char*    msg)
{
   ofstream    fOut(fspec, ios::app);
   fOut << msg;
}
//****************************************************************************
void
LoadCtext(
   char* fspec)
{
                                       // read in the ciphertext
   if ((fpCtext = fopen(fspecCtext, "rb")) == NULL)
   {
      sprintf(vcMsgBuf,
         "error opening %s for input.\n", fspecCtext);
      exit_w_msg(vcMsgBuf);
   }
                                       /* get the file length minus the IV */
   fseek(fpCtext, 0L, SEEK_END);
   lenCtext = ftell(fpCtext) - 8;
   rewind(fpCtext);
                                       /* read the IV and ciphertext */
   if (fread(IV, 1, 8, fpCtext) != 8)
   {
      sprintf(vcMsgBuf,
         "error reading from %s.\n", fspecCtext);
      exit_w_msg(vcMsgBuf);
   }

   if (  (ciphertext = (Uchar*)malloc(lenCtext)) == NULL
      || (plaintext  = (Uchar*)malloc(lenCtext)) == NULL   )
   {
      exit_w_msg("error allocating memory.\n");
   }
   if (fread(ciphertext, 1, lenCtext, fpCtext) != size_t(lenCtext))
   {
      sprintf(vcMsgBuf,
         "error reading from %s.\n", fspecCtext);
      exit_w_msg(vcMsgBuf);
   }
   fclose(fpCtext);
}
//****************************************************************************
void
FreeCtext(void)
{
   if (ciphertext)   free(ciphertext);
   if (plaintext)    free(plaintext);
}
//****************************************************************************
                                       // for console app
#if defined(BRUTECON)

int
main(
   int      argc,
   char**   argv)
{
   short    ctCyc = sizeof(cyctable);


   printf("Brute v1.0 - Oct 1997 Counterpane Systems, SynData Technologies\n");

                                       // read command input
   GetState(fspecState);
   LoadCtext(fspecCtext);

                                       // set Ctrl-C signal handler
   signal(SIGINT, (fnSignal) wrapup);


                                       // loop through the cycles

   short iStartCycle = iCurrCycle;     // first, probably partial, cycle
                                       //    as specified by input
   printf("\nRestart point - cycle %d, key %s",
      iCurrCycle,
      hexNextKey);
   (void) search(NextKey, cyctable[iStartCycle].DelimKey);

   for ( iCurrCycle++;
         iCurrCycle!=iStartCycle;
         iCurrCycle++, iCurrCycle%=28)
   {
      printf("\nCycle %d - %lu chunk(s)",
            iCurrCycle,
            cyctable[iCurrCycle].uCycLenHi + 1);
      memcpy(NextKey, cyctable[iCurrCycle].DelimKey, 5);
      search(NextKey, cyctable[iCurrCycle].DelimKey);
   }
   printf("\n");
   FreeCtext();
   exit(0);
}

#endif   // defined(BRUTECON)

//****************************************************************************
                                       // for screen saver app
#if defined(BRUTESAV)

void
search_one_chunk(void)
{

   char  windir[MAX_PATH];

   GetWindowsDirectory(windir, MAX_PATH);
   _chdir(windir);

   GetState(fspecState);

   if (bEnabled)
      LoadCtext(fspecCtext);

   if (bEnabled)
   {
      sprintf(vcMsgBuf,
         "Processing cycle %d\n", iCurrCycle);
      if (0 == search(NextKey, cyctable[iCurrCycle].DelimKey, k2to20th))
      {
         iCurrCycle++;
         iCurrCycle%=28;
         memcpy(NextKey, cyctable[iCurrCycle].DelimKey, 5);
         PutState(fspecState);
      }
   }

   FreeCtext();
}
#endif   // defined(BRUTESAV)
//****************************************************************************
//****************************************************************************


